//-----------------------------------------------------------------------------
// Written by xawari. Copyright (c) 2018
// Distributed under the terms of Mozilla Public License version 2.0
//-----------------------------------------------------------------------------
#include "platform.h"
#include "color.h"
#include <stddef.h>

#define NUM_PALETTE_COLORS 256u// we could rewrite program to use dynamic value, but nobody needs that anyway

//int g_PrintLevel = 0;
typedef int (__cdecl *comparefunc_t)(const void *, const void *);

typedef struct unicolor_s
{
	Color c;
	float h,s,l;// keep converted color associated with original RGB
} unicolor_t;

unicolor_t g_ColorTable[NUM_PALETTE_COLORS];




#if defined (_DEBUG)
//-----------------------------------------------------------------------------
// Purpose: Check validity of expression and raise an error message on failure
//-----------------------------------------------------------------------------
bool DBG_AssertFunction(bool fExpr, const char *szExpr, const char *szFile, int iLine, const char *szMessage)
{
	if (fExpr)
		return true;

	if (szMessage != nullptr)
		printf("ASSERT FAILED: %s (%s@%d)\n%s", szExpr, szFile, iLine, szMessage);
	else
		printf("ASSERT FAILED: %s (%s@%d)\n", szExpr, szFile, iLine);

#if _MSC_VER > 1200
	if (_CrtDbgReport(_CRT_ASSERT, szFile, iLine, nullptr, szExpr) == 1)// "retry"
		_CrtDbgBreak();
#else// older VS and other platforms
#if defined (_WIN32)
		_assert((void *)szExpr, (void *)szFile, iLine);
#else// if there's no POSIX/ISO equivalent for this:
		fprintf(stderr, "ASSERT FAILED: %s (%s@%d)\n", szExpr, szFile, iLine);
		//getch();// halt execuiton?
#endif
#endif
	return fExpr;
}
#else// XDM: ASSERT function for release build
bool NDB_AssertFunction(bool fExpr, const char *szExpr, const char *szMessage)
{
	if (fExpr)
		return true;

	if (szMessage != nullptr)
		printf("ASSERT FAILED: %s %s\n", szExpr, szMessage);
	else
		printf("ASSERT FAILED: %s\n", szExpr);

	return fExpr;
}
#endif	// DEBUG

void NormalizeValueF(float *value, const float &min, const float &max)
{
	float range = max-min;
	if (*value >= max)// XDM3039: to match the debug version algorithm
		*value -= range;
	else if (*value < min)
		*value += range;
}

float NormalizeValueFr(const float &srcvalue, const float &min, const float &max)
{
	static float fRet;
	fRet = srcvalue;
	NormalizeValueF(&fRet, min, max);
	return fRet;
}

void NormalizeAngle360(float *angle)
{
	NormalizeValueF(angle, 0, 360);
}
float NormalizeAngle360r(const float &angle)
{
	return NormalizeValueFr(angle, 0, 360);
}



//-----------------------------------------------------------------------------
// Purpose: Does this file path end with specified extension?
// Input  : *path - absolute/relative path to a file
//			*checkext - ".ext" with dot!
// Output : Returns true on success, false on failure.
//-----------------------------------------------------------------------------
bool UTIL_FileExtensionIs(const char *path, const char *checkext)
{
	if (path == nullptr || checkext == nullptr)
		return false;

	size_t len = strlen(path);
	size_t extlen = strlen(checkext);
	if (len < 2 || extlen < 1 || len < extlen)
		return false;

	bool bFail = false;
	do
	{
		--len;
		--extlen;
		if (path[len] != checkext[extlen])
		{
			bFail = true;
			break;
		}
	}
	while (len > 0 && extlen > 0);

	if (bFail == false)
	{
// if there was no dot in *checkext		if (path[len-1] == '.')
			return true;
	}
	return false;
}

//-----------------------------------------------------------------------------
// Purpose: Parse 256 RGB bytes from a binary file (ripped out of XDM)
// Input  : *filename - palette.lmp
// Output : Returns true on success, false on failure.
//-----------------------------------------------------------------------------
bool UTIL_LoadRawPalette(const char *filename)
{
	// DEBUG printf("LoadRawPalette(%s)\n", filename);
	FILE *pFile = fopen(filename, "rb");
	if (pFile == nullptr)
		return false;

	// NO! fread(g_Palette, 3, 256, pFile);// Color is RGBA, thus 4 bytes, but we only have RGB in palette
	size_t n;
	byte rgb[3];
	for (size_t i = 0; i<NUM_PALETTE_COLORS; ++i)
	{
		n = fread(rgb, 1, 3, pFile);
		if (n < 3)
		{
			printf("LoadRawPalette(%s) error: failed to read RGB data at offset %d!\n", filename, i*3+n);
			return false;
		}
		g_ColorTable[i].c.Set3b(rgb);
		// debug conprintf(2, "LoadRawPalette(%s): %08X\n", filename, g_Palette[i].integer);
	}
	return true;
}


//-----------------------------------------------------------------------------
// Purpose: Parse 256 RGB bytes from a JASC pal file (text)
// Input  : *pFileName - palette.pal
// Output : Returns true on success, false on failure.
//-----------------------------------------------------------------------------
bool ReadPalFile(const char *pFileName)
{
	// DEBUG printf("ReadPalFile(%s)\n", pFileName);
	FILE *palfile = fopen(pFileName, "rt");
	if (palfile == nullptr)
		return false;

	char cur_str[16];
	if (fgets(cur_str, 16, palfile) == nullptr)
		printf("fgets error!\n");
	else if(strcmp(cur_str, "JASC-PAL\n") != 0)
	{
		printf("Invalid palette type: %s\n", cur_str);
		fclose(palfile);
		return false;
	}

	fgets(cur_str, 16, palfile);// skip "0100" line
	//printf("%s\n", cur_str);
	int i;
	fgets(cur_str, 16, palfile);// read number of colors
	i = atoi(cur_str);
	if (i != NUM_PALETTE_COLORS)// UNDONE: any number of colors
	{
		printf("Unsupported number of colors: %d\n", i);
		fclose(palfile);
		return false;
	}

	i = 0;
	while (fgets(cur_str, 16, palfile))
	{
		//printf("Reading color %d: %s", i, cur_str);// don't use \n! fgets already did it!
		if (!StringToRGB(cur_str, g_ColorTable[i].c.r, g_ColorTable[i].c.g, g_ColorTable[i].c.b))
			printf("Error reading color entry %d!\n", i);

		++i;
	}
	fclose(palfile);
	return true;
}

//-----------------------------------------------------------------------------
// Purpose: Write 256 RGB bytes to a JASC pal file (text)
// Input  : *pFileName - palette.pal
// Output : Returns true on success, false on failure.
//-----------------------------------------------------------------------------
bool WritePalFile(const char *pFileName)
{
	printf("WritePalFile(%s)\n", pFileName);
	FILE *pFile = fopen(pFileName, "wt");
	if (pFile == nullptr)
		return false;

	fprintf(pFile, "JASC-PAL\n0100\n%u\n", NUM_PALETTE_COLORS);// write the header // UNDONE: any number of colors
	for (size_t i=0; i < NUM_PALETTE_COLORS; ++i)
	{
		if (fprintf(pFile, "%hu %hu %hu\n", (uint16)g_ColorTable[i].c.r, (uint16)g_ColorTable[i].c.g, (uint16)g_ColorTable[i].c.b) < 0)
		{
			printf("Error writing color entry %u!\n", i);
			break;
		}
	}
	fclose(pFile);
	return true;
}

//-----------------------------------------------------------------------------
// Purpose: Write 256 RGB bytes to a binary file (binary)
// Input  : *pFileName - palette.act
//			actfile - write Adobe(tm) Color Table format, which adds 2 shorts: number of colors and transparent color index.
// Output : Returns true on success, false on failure.
//-----------------------------------------------------------------------------
bool WriteBinFile(const char *pFileName, bool actfile)
{
	printf("WriteBinFile(%s, %s)\n", pFileName, actfile?"ACT":"RAW");
	FILE *pFile = fopen(pFileName, "wb");
	if (pFile == nullptr)
		return false;

	for (size_t i=0; i < NUM_PALETTE_COLORS; ++i)
	{
		if (fwrite(&g_ColorTable[i].c, sizeof(byte)*3, 1, pFile) < 3)// unfortunately, we cannot just dump all data at once because it's in RGBA format
		{
			printf("Error writing color entry %u!\n", i);
			break;
		}
	}
	if (actfile)
	{
		uint16 data = NUM_PALETTE_COLORS;// number of colors
		// DEBUG printf("Writing n colors %hu\n", data);
		data = _byteswap_ushort(data);
		fwrite(&data, sizeof(uint16), 1, pFile);
		data = NUM_PALETTE_COLORS-1;// last color is for transparency
		// DEBUG printf("Writing trans color %hu\n", data);
		data = _byteswap_ushort(data);
		fwrite(&data, sizeof(uint16), 1, pFile);
	}
	fclose(pFile);
	return true;
}

//-----------------------------------------------------------------------------
// Purpose: Sort array of colors
// Note   : The array is sorted in increasing order, as defined by the comparison function.
// Note   : So return 1 to push elem1 to the end of the list
// Input  : *elem1 - 
//			*elem2 - 
// Output : int <0, 0, >0 : elem1 less than, equivalent to, greater than elem2
//-----------------------------------------------------------------------------
int sort_h(const void *elem1, const void *elem2)
{
	if (((const unicolor_t *)elem1)->h < ((const unicolor_t *)elem2)->h)
		return -1;
	else if (((const unicolor_t *)elem1)->h > ((const unicolor_t *)elem2)->h)
		return 1;
	return 0;
}

int sort_s(const void *elem1, const void *elem2)
{
	if (((const unicolor_t *)elem1)->c.integer == 0 || ((const unicolor_t *)elem2)->c.integer == 0 ||
		((const unicolor_t *)elem1)->c.integer == 0xFFFFFF00 || ((const unicolor_t *)elem2)->c.integer == 0xFFFFFF00)
		return 0;// saturation is undetermined, do not move

	if (((const unicolor_t *)elem1)->s < ((const unicolor_t *)elem2)->s)
		return -1;
	else if (((const unicolor_t *)elem1)->s > ((const unicolor_t *)elem2)->s)
		return 1;
	return 0;
}

int sort_l(const void *elem1, const void *elem2)
{
	if (((const unicolor_t *)elem1)->l < ((const unicolor_t *)elem2)->l)
		return -1;
	else if (((const unicolor_t *)elem1)->l > ((const unicolor_t *)elem2)->l)
		return 1;
	return 0;
}
int sort_r(const void *elem1, const void *elem2)
{
	if (((const unicolor_t *)elem1)->c.r < ((const unicolor_t *)elem2)->c.r)
		return -1;
	else if (((const unicolor_t *)elem1)->c.r > ((const unicolor_t *)elem2)->c.r)
		return 1;
	return 0;
}

int sort_g(const void *elem1, const void *elem2)
{
	if (((const unicolor_t *)elem1)->c.g < ((const unicolor_t *)elem2)->c.g)
		return -1;
	else if (((const unicolor_t *)elem1)->c.g > ((const unicolor_t *)elem2)->c.g)
		return 1;
	return 0;
}

int sort_b(const void *elem1, const void *elem2)
{
	if (((unicolor_t *)elem1)->c.b < ((unicolor_t *)elem2)->c.b)
		return -1;
	else if (((unicolor_t *)elem1)->c.b > ((unicolor_t *)elem2)->c.b)
		return 1;
	return 0;
}

int sort_lum(const void *elem1, const void *elem2)
{
	unicolor_t *pc1 = (unicolor_t *)elem1;
	unicolor_t *pc2 = (unicolor_t *)elem2;
	float lum1 = (pc1->c.r*pc1->c.r + pc1->c.g*pc1->c.g + pc1->c.b*pc1->c.b);// sqrt() can be omitted for comparison
	float lum2 = (pc2->c.r*pc2->c.r + pc2->c.g*pc2->c.g + pc2->c.b*pc2->c.b);
	if (lum1 < lum2)
		return -1;
	else if (lum1 > lum2)
		return 1;
	return 0;
}

//-----------------------------------------------------------------------------
// Purpose: Parse range out of a string
// Input  : pText - range in "a-b" format or just "all"
//			rangestart - output
//			rangeend - output
//-----------------------------------------------------------------------------
bool ParseRange(const char *pText, size_t &rangestart, size_t &rangeend)
{
	if (pText)
	{
		if (sscanf(pText, "%u-%u", &rangestart, &rangeend) == 2)
		{
			if (rangestart < rangeend)
				return true;
		}
		else if (strcmp(pText, "all") == 0)
		{
			rangestart = 0;
			rangeend = NUM_PALETTE_COLORS-1;
			return true;
		}
	}
	return false;
}

//-----------------------------------------------------------------------------
// Purpose: Fix the color data
// Input  : c - color
// h - Hue is 0...360
// s - Saturation is 0...1
// l - Lightness is 0...1
//-----------------------------------------------------------------------------
void ValidateColor(unicolor_t &c)
{
	// RGB are bytes, so are warped around 255 with no way of detecting an error
	NormalizeValueF(&c.h, 0, 360);
	c.s = clamp(c.s, 0, 1);
	c.l = clamp(c.l, 0, 1);
}


void showusage(void)
{
	printf("Usage: programname <infile> <outfile> [--out-raw] [--out-std] [--out-act] [reverse | print | sort <range> <mode> | mod <range> <component> <operator> <value>] ...\n range is \"start-end\"|\"all\"; component or sort mode is r|g|b|h|s|l|t; operator is =|+=|-=|*=|/=;\n Commands are executed in specified order.\n Only %u color JASC or RAW RGB palette files are supported.\n", NUM_PALETTE_COLORS);
}

//-----------------------------------------------------------------------------
// Purpose: Main
//-----------------------------------------------------------------------------
int main(int argc, char **argv)
{
	printf("\n - Palette processing tool 1.1 (build %s) by xawari -\n", __DATE__);
	if (argc <= 1)
	{
		showusage();
		return 1;
	}
#if defined (_DEBUG)
//	g_PrintLevel = 2;
#endif
	// Validate I/O arguments
	char *pInputName = nullptr;
	char *pOutputName = nullptr;

	if (strcmp(argv[1], "-h") == 0 || strcmp(argv[1], "--help") == 0)
	{
		showusage();
		return 0;// this was expected
	}
	else
		pInputName = argv[1];

	if (!pInputName)
	{
		printf("Error: no input name.\n");
		return 1;
	}

	pOutputName = argv[2];
	if (!pOutputName)
	{
		printf("Error: no output name.\n");
		return 1;
	}
	if (strcmp(pInputName, pOutputName) == 0)
	{
		printf("Warning: input and output file names are the same!\n");
		//return 1;
	}

	int iErrorMode = 0;
	int option = 3;
	bool bOutputBinary = false;
	bool bOutputAct = false;
	bool bOutputStd = false;
	for (; option < argc; ++option)
	{
		//DEBUG printf(" option %d: %s\n", option, argv[option]);
		if (argv[option][0] != '-')// options only
			break;

		if (strcmp(argv[option], "--out-raw") == 0)
			bOutputBinary = true;
		else if (strcmp(argv[option], "--out-act") == 0)
		{
			bOutputBinary = true;
			bOutputAct = true;
		}
		else if (strcmp(argv[option], "--out-std") == 0)
			bOutputStd = true;
		else
			printf("Unknown option: %s\n", argv[option]);
	}
	//DEBUG printf("Commands start from: %d\n", option);

	memset(&g_ColorTable, 0, sizeof(g_ColorTable[0])*NUM_PALETTE_COLORS);

	printf("Input: \"%s\"\n", pInputName);
	if (UTIL_FileExtensionIs(pInputName, ".pal"))
	{
		if (!ReadPalFile(pInputName))
			iErrorMode = 1;
	}
	else
	{
		if (!UTIL_LoadRawPalette(pInputName))
			iErrorMode = 1;
	}

	if (iErrorMode != 0)
	{
		printf("Error reading input file \"%s\"!\n", pInputName);
		return iErrorMode;
	}

	// Initialize HSL color model
	for (uint i=0; i < NUM_PALETTE_COLORS; ++i)
	{
		g_ColorTable[i].h = 1; g_ColorTable[i].s = 1; g_ColorTable[i].l = 1;
		RGB2HSL(g_ColorTable[i].c.r, g_ColorTable[i].c.g, g_ColorTable[i].c.b, g_ColorTable[i].h, g_ColorTable[i].s, g_ColorTable[i].l);
	}

	comparefunc_t CompareFunc = nullptr;
	const char *act;
	for (; option < argc; ++option)
	{
		act = argv[option];
		if (act == nullptr || *act == 0)
		{
			option = argc;
			//break;
		}
		else if (strcmp(act, "reverse") == 0 || strcmp(act, "rev") == 0)
		{
			if (argc > option + 1)// reverse 0-127
			{
				option++;
				size_t rangestart = 0, rangeend = NUM_PALETTE_COLORS-1;
				if (!ParseRange(argv[option], rangestart, rangeend))
				{
					printf(" - %s: error: bad range %s!\n", act, argv[option+1]);
					iErrorMode = 1;
					break;
				}
				size_t halfrange = (rangeend - (rangeend-rangestart+1)/2);
				printf(" - %s: processing range of %u colors %u-%u...\n", act, rangeend-rangestart+1, rangestart, rangeend);
				unicolor_t tmp;
				for (uint i=rangestart; i < halfrange; ++i)// NUM_PALETTE_COLORS/2
					SWAP(g_ColorTable[i], g_ColorTable[rangeend-i], tmp);
			}
			else
			{
				printf(" - %s: error: not enough arguments!\n", act);
				iErrorMode = 1;
				break;
			}
		}
		else if (strcmp(act, "print") == 0)
		{
			for (uint i=0; i < NUM_PALETTE_COLORS; ++i)
				printf("%03u %u %u %u %u %f %f %f\n", i, g_ColorTable[i].c.integer, (uint)g_ColorTable[i].c.r, (uint)g_ColorTable[i].c.g, (uint)g_ColorTable[i].c.b, g_ColorTable[i].h, g_ColorTable[i].s, g_ColorTable[i].l);
		}
		else if (strcmp(act, "sort") == 0)
		{
			if (argc > option + 2)// sort 0-127 h
			{
				option++;
				size_t rangestart = 0, rangeend = NUM_PALETTE_COLORS-1;
				if (!ParseRange(argv[option], rangestart, rangeend))
				{
					printf(" - %s: error: bad range %s!\n", act, argv[option]);
					iErrorMode = 1;
					break;
				}
				rangestart = min(rangestart, NUM_PALETTE_COLORS-2);
				rangeend = min(rangeend, NUM_PALETTE_COLORS-1);

				option++;
				const char *pSortMode = argv[option];// TODO: accept combination of letters? e.g. "ht"
				if (pSortMode[0] == 'h')
					CompareFunc = sort_h;
				else if (pSortMode[0] == 's')
					CompareFunc = sort_s;
				else if (pSortMode[0] == 'l')
					CompareFunc = sort_l;
				else if (pSortMode[0] == 'r')
					CompareFunc = sort_r;
				else if (pSortMode[0] == 'g')
					CompareFunc = sort_g;
				else if (pSortMode[0] == 'b')
					CompareFunc = sort_b;
				else if (pSortMode[0] == 't')
					CompareFunc = sort_lum;
				else
				{
					CompareFunc = nullptr;
					printf(" - %s: error: unknown mode %s!\n", act, pSortMode);
					iErrorMode = 1;
					break;
				}
				if (CompareFunc)
				{
					printf(" - %s: processing range of %u colors %u-%u %s...\n", act, rangeend-rangestart+1, rangestart, rangeend, pSortMode);
					qsort(&g_ColorTable[rangestart], rangeend-rangestart+1, sizeof(g_ColorTable[0]), CompareFunc);
				}
			}
			else
			{
				printf(" - %s: error: not enough arguments!\n", act);
				iErrorMode = 1;
				break;
			}
		}
		else if (strcmp(act, "mod") == 0)
		{
			if (argc > option + 4)// mod 0-239 r += 20
			{
				option++;
				size_t iDataOffset = 0, rangestart = 0, rangeend = NUM_PALETTE_COLORS-1;
				if (!ParseRange(argv[option], rangestart, rangeend))
				{
					printf(" - %s: error: bad range %s!\n", act, argv[option]);
					iErrorMode = 1;
					break;
				}

				option++;
				const char *pTargetName = argv[option];
				bool bUseHSL = false;
				if (pTargetName[0] == 'h')
				{
					iDataOffset = offsetof(unicolor_s, h);
					bUseHSL = true;
				}
				else if (pTargetName[0] == 's')
				{
					iDataOffset = offsetof(unicolor_s, s);
					bUseHSL = true;
				}
				else if (pTargetName[0] == 'l')
				{
					iDataOffset = offsetof(unicolor_s, l);
					bUseHSL = true;
				}
				else if (pTargetName[0] == 'r')
					iDataOffset = offsetof(unicolor_s, c.r);
				else if (pTargetName[0] == 'g')
					iDataOffset = offsetof(unicolor_s, c.g);
				else if (pTargetName[0] == 'b')
					iDataOffset = offsetof(unicolor_s, c.b);
				else
				{
					printf(" - %s: error: unknown component %s!\n", act, pTargetName);
					iErrorMode = 1;
					break;
				}

				option++;
				float *pTargetHSL = nullptr;
				byte *pTargetRGB = nullptr;
				const char *pOperator = argv[option];
				option++;
				float value = atof(argv[option]);
				printf(" - %s: modifying range of %u colors %u-%u %s %s %g...\n", act, rangeend-rangestart+1, rangestart, rangeend, pTargetName, pOperator, value);
				for (size_t i = rangestart; i <= rangeend; ++i)
				{
					if (bUseHSL)
						pTargetHSL = (float *)((byte *)&g_ColorTable[i] + iDataOffset);
					else
						pTargetRGB = (byte *)&g_ColorTable[i] + iDataOffset;

					// Add operators here
					if (strcmp(pOperator, "=") == 0)
					{
						if (pTargetHSL)
							*pTargetHSL = value;
						else if (pTargetRGB)
							*pTargetRGB = (int)value;
					}
					else if (strcmp(pOperator, "+=") == 0)
					{
						if (pTargetHSL)
							*pTargetHSL += value;
						else if (pTargetRGB)
							*pTargetRGB += (int)value;
					}
					else if (strcmp(pOperator, "-=") == 0)
					{
						if (pTargetHSL)
							*pTargetHSL -= value;
						else if (pTargetRGB)
							*pTargetRGB -= (int)value;
					}
					else if (strcmp(pOperator, "*=") == 0)
					{
						if (pTargetHSL)
							*pTargetHSL *= value;
						else if (pTargetRGB)
							*pTargetRGB *= (int)value;
					}
					else if (strcmp(pOperator, "/=") == 0)
					{
						if (pTargetHSL)
							*pTargetHSL /= value;
						else if (pTargetRGB)
							*pTargetRGB /= (int)value;
					}
					else
					{
						printf(" - %s: error: unknown operator: \"%s\"\n", act, pOperator);
						iErrorMode = 1;
						break;
					}

					ValidateColor(g_ColorTable[i]);// fix all ipossible valuse
					// now synchronize all color components
					if (bUseHSL)// we modified HSL, so translate it to RGB
						HSL2RGB(g_ColorTable[i].h, g_ColorTable[i].s, g_ColorTable[i].l, g_ColorTable[i].c.r, g_ColorTable[i].c.g, g_ColorTable[i].c.b);
					else// we modified RGB, so translate it to HSL
						RGB2HSL(g_ColorTable[i].c.r, g_ColorTable[i].c.g, g_ColorTable[i].c.b, g_ColorTable[i].h, g_ColorTable[i].s, g_ColorTable[i].l);
				}
			}
			else
			{
				printf(" - %s: error: not enough arguments!\n", act);
				iErrorMode = 1;
				break;
			}
		}
		else
			printf("Unknown command: %s\n", argv[option]);

		if (iErrorMode > 0)
			break;
	}

	if (iErrorMode == 0)
	{
		if (bOutputStd)// to console
		{
			if (bOutputBinary)
			{
				for (size_t i=0; i < NUM_PALETTE_COLORS; ++i)
				{
					if (fwrite(&g_ColorTable[i].c, 3, 1, stdout) == 0)// only RGB
					{
						iErrorMode = 1;
						break;
					}
				}
			}
			else
			{
				printf("INDEX8 RGBA32 R8 G8 B8 HF32 SF32 LF32\n");
				for (uint i=0; i < NUM_PALETTE_COLORS; ++i)
					printf("%03u %u  %u %u %u %f %f %f\n", i, g_ColorTable[i].c.integer, (uint)g_ColorTable[i].c.r, (uint)g_ColorTable[i].c.g, (uint)g_ColorTable[i].c.b, g_ColorTable[i].h, g_ColorTable[i].s, g_ColorTable[i].l);
			}
		}
		else
		{
			printf("Output: \"%s\"\n", pOutputName);
			if (bOutputBinary || bOutputAct)
			{
				if (!WriteBinFile(pOutputName, bOutputAct))
					iErrorMode = 1;
			}
			else
			{
				if (!WritePalFile(pOutputName))
					iErrorMode = 1;
			}
		}
	}
	else
		printf("Errors detected, not writing output.\n");

//finish:
	return iErrorMode;
}
